#!/usr/bin/perl 
#
# Copyright (c) 2005  Gregor Maier <gregor@majordomus.org>
# All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. Neither the names of the copyright owners nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
# 4. All advertising materials mentioning features or use of this software
#    must display the following acknowledgement:
#        This product includes software developed by Anzen Computing, Inc.
# 
# THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
# GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
# IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
# OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
# ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# 

use strict; 
use POSIX;

########### CONFIG ##################

# WARNING: You should really use an absolut path here
# the relative path is for the sake of example 

# Directory where the output files should be places
my $OUTDIR="./addrconf/output";

# Directory where copies of the output files and the 
# config should be placed
my $BAKDIR="./addrconf/backup";

# Inputfile where all the hosts are
# specified.
my $INPUT="./addrconf/addrconf.conf";


#
# The Domainname to append to hostname for the reverse zone. 
# The last dot for the root zone is added automaticly
my $DOMAIN="example.com";

# Filename for the Forward zone. 
# NOTE: This file _only_ contains A and AAAA records. CNAMES, NS, MX 
# and other records are not handled by this file
my $DNS_v4_FILENAME="entries.hosts.v4";
my $DNS_v6_FILENAME="entries.hosts.v6";

# Filename for the v6 reverse zone 
# This file will only contain PTR records. SOA must
# be added externally
my $DNS_v6_REV_FILENAME="entries.ip6.arpa";


# The IPv4 reverse zone, that we might encounter
# i.e. multiple v4 reverse zone can be generated
# Each generated zone assumes an ORIGIN of 3 octets
#
my @IPV4_REV_ZONES = ('42.168.192.in-addr.arpa', '23.168.192.in-addr.arpa');
# The filenames for the reverse IPv4 zone entries are generated as:
#    $IPV4_REV_FILENAME_PREFIX$IPV4_REV_ZONES[i]
#    eg. entries.42.168.192.in-addr.arpa
my $IPV4_REV_FILENAME_PREFIX="entries.";

# The length of the IPv6 Prefix for the v6 reverse zone 
# Only one IPv6 reverse zone file will be generated, with
# an expected $ORIGIN of $PREFIXLEN bit
my $PREFIXLEN=48;


# Filenames for the dhcp config files
# A seperate filename for each "dhcp-realm" that can be encountered
# must be speicifed here
my %DHCP_FILENAMES;
$DHCP_FILENAMES{'workstations'} = "workstations.dhcpd.conf";
$DHCP_FILENAMES{'dmz'} = "dmz.dhcpd.conf";

################ CONFIG ENDS ###########################

# There shouldn't be the need to change anything beyond this line. 


########### GLOBAL VARS ##############

# filehandles
my $DNSFW4; 		# forward zone handle
my $DNSFW6; 		# forward zone handle
my %DNSREV4;	# IPv4 reverse zone handles
my $DNSREV6;	# IPv6 reverse zone handle
my %DHCP_CONF;


my $errors=0;


# all flags we have:
my $dhcp;
my $NOREV;

&main;


sub main {
	my ($host, $mac, $ipv4, $ipv6type, $ipv6, $flagstr, $comment);
	my $ipv6addr;
	my @elements;

	open IN, "< $INPUT" or 
			die "FATAL: Could not open $INPUT for reading!\n";

	&do_backup;

	if (!-d $OUTDIR) { 
		die "FATAL: Basedirectory \"$OUTDIR\"  does not exist!\n";
	}

	my @fhandles;
	open $DNSFW4, "> $OUTDIR/$DNS_v4_FILENAME" or 
			die "FATAL: Could not open $OUTDIR/$DNS_v4_FILENAME for writing\n";
	push @fhandles, $DNSFW4;

	open $DNSFW6, "> $OUTDIR/$DNS_v6_FILENAME" or 
			die "FATAL: Could not open $OUTDIR/$DNS_v6_FILENAME for writing\n";
	push @fhandles, $DNSFW6;

	open $DNSREV6, "> $OUTDIR/$DNS_v6_REV_FILENAME" or 
			die "FATAL: Could not open $OUTDIR/$DNS_v6_REV_FILENAME for writing\n";
	push @fhandles, $DNSREV6;

	foreach (@IPV4_REV_ZONES) {
		open $DNSREV4{$_}, "> $OUTDIR/$IPV4_REV_FILENAME_PREFIX$_" or
			die "FATAL: Could not open $OUTDIR/$IPV4_REV_FILENAME_PREFIX$_ for writing\n";
		push @fhandles, $DNSREV4{$_};
	}

	foreach (@fhandles) {
		print $_ "; AUTOMATICLY GENERATED FILE -- DO NOT EDIT\n";
		print $_ "; I REALLY MEAN DO NOT EDIT!!\n";
	}

	foreach (keys %DHCP_FILENAMES) {
		open $DHCP_CONF{$_},  "> $OUTDIR/$DHCP_FILENAMES{$_}" or
			die "FATAL: Could not open $OUTDIR/" . $DHCP_FILENAMES{$_} . " for writing\n";
		my $tmpfh = $DHCP_CONF{$_};
		print $tmpfh "# AUTOMATICLY GENERATED FILE -- DO NOT EDIT\n";
		print $tmpfh "# I REALLY MEAN DO NOT EDIT!!\n";
	}


	while (<IN>) {
		chomp;
		next if /^\s*#/;   # discard comments
		next if /^\s*$/;   # discard empty lines

		# $_ =~ s/(.*)#.*/\1/;  # Remove commets

		@elements = split /\|/, $_, 7; # Split line into tokens
		# Remoe leading and trailing whitespaces for each token
		foreach (@elements) {

			 s/^\s*(\S*)\s*$/\1/;
		}
		($host, $mac, $ipv4, $ipv6type, $ipv6, $flagstr, $comment) = @elements;

		# Split the flags 
		my @flagsarr = undef;
		@flagsarr = split /,/, $flagstr;
		$dhcp = "";
		$NOREV=0;
		foreach my $curflag (@flagsarr) {
			# Remove leading and trailing whitespaces
			$curflag =~ s/^\s*(\S*)\s*$/\1/;

			
			# If the flag is all lowercase, it's the dhcp-realm
			if ($curflag =~ /^[a-z][a-z0-9]*$/) {
				$dhcp=$curflag;
			}
			elsif ($curflag eq "NOREV") {
				$NOREV="1";
			}
			else { # unknown Flag
				print STDERR "ERROR: flag \"$curflag\" for host $host is unknown\n";
				$errors++;
			}
		} # end split flags

		$ipv6addr = get_ipv6_addr($ipv6type, $ipv6, $mac, $host);
		print_dhcp($host, $mac, $dhcp);
		print_dns_v4_forward($host,$ipv4);
		print_dns_v6_forward($host,$ipv6addr);
		print_dns_v4_rev($host, $ipv4);
		print_dns_v6_rev($host, $ipv6addr);

	}
	foreach (@fhandles) {
		close $_;
	}
	foreach (values %DHCP_CONF) {
		close $_
	}
} # end main


sub do_backup {
	if (! -d $BAKDIR) {
		die "FATAL: Backup directory $BAKDIR does not exist";
	}

	my @ts=localtime(time());	
	my $datestr = strftime("%Y%m%d--%H%M%S", @ts);

	system("mkdir $BAKDIR/$datestr");
	system("mv $OUTDIR/* $BAKDIR/$datestr/");
	system("cp $INPUT $BAKDIR/$datestr/");
}

sub get_ipv6_addr {
	my ($ipv6type, $ipv6, $mac, $host) = @_;

	my $ipv6addr="";
	if ($ipv6type eq "fix") {
		$ipv6addr = $ipv6;
	}
	elsif ($ipv6type eq "auto") {
		my $tmpaddr = `ipv6calc --in mac $mac --action geneui64 --out eui64`;
		chomp $tmpaddr;
		if ($tmpaddr =~ /[0-9a-fA-F:]+/) {
			$ipv6addr = "$ipv6:$tmpaddr";
		}
		else {
			print STDERR "ERROR: ipv6calc generated stupid results for host $host\n";
			$errors++;
		}
	}
	elsif ($ipv6type eq "none") {
		; # do nothing
	}
	else {
		print STDERR "ERROR: ipv6 type \"$ipv6type\" for host $host is unknown\n";
		$errors++;
	}
}

# DNS FORWARD ZONE
sub print_dns_v4_forward ($$) {
	my ($host, $ipv4) = @_;

	print $DNSFW4 "$host\tIN A\t$ipv4\n";
}

sub print_dns_v6_forward ($$) {
	my ($host, $ipv6addr) = @_;

	if ($ipv6addr ne "") {
		print $DNSFW6 "$host\tIN AAAA\t$ipv6addr\n";
	}
}

sub print_dns_v4_rev ($$){
	my ($host, $ipv4) = @_;
	# DNS IPv4 REVERSE ZONE(S)
	my @octets;
	my $revzone;
	if (!$NOREV) { 
		@octets = split /\./, $ipv4;
		$revzone = "$octets[2].$octets[1].$octets[0].in-addr.arpa";
		my $fh=$DNSREV4{$revzone};
		print $fh "$octets[3]\tIN PTR\t$host.$DOMAIN.\n";
	}
	else {
		print STDERR "NOTE: no reverse IPv4 for $host\n";
	}
}

sub print_dns_v6_rev ($$) { 
	my ($host, $ipv6addr) = @_;


	# DNS IPv6 REVERSE ZONE
	my $revaddr;
	if (!$NOREV && ($ipv6addr ne "")) { 
		$revaddr = `ipv6calc --in ipv6addr $ipv6addr --out revnibbles.arpa`;
		chomp $revaddr;
		if ($revaddr =~ /^([a-fA-F0-9]\.){32}ip6.arpa.$/ ) {
				# each 4 Bits of the address is 2 characters wide (nibble + dot)
				# therefor: the number of nibbles we need is  (128-prefixlen)/4
				# then multiply with two to account for the dot 
				# then subtract 1 to get rid of the trailing dot.
				$revaddr = substr($revaddr, 0, (128-$PREFIXLEN)/2-1); 
				print $DNSREV6 "$revaddr\tIN PTR\t$host.$DOMAIN.\n";
		}
		else { # ipv6calc returned bogus data
			print STDERR "ERROR: ipv6calc generated stupid results for host $host (reverse entry)\n";
			$errors++;
		}
		
	}
}


sub print_dhcp ($$$) {
	my ($host, $mac, $dhcp) = @_;
	
	if ($dhcp ne "") {
		if ($DHCP_CONF{$dhcp} == undef) {
			print STDERR "ERROR: no dhcp output file for dhcp realm \"$dhcp\" defined. Host was $host\n";
			$errors++;
			return;
		}
		my $tmpfh = $DHCP_CONF{$dhcp};
		print $tmpfh "host $host {\n";
		print $tmpfh "    hardware ethernet $mac;\n";
		print $tmpfh "    fixed-address $host.$DOMAIN;\n";
		print $tmpfh "}\n";
	}
}
